//	xExECZT[ BME280(I2C W[ GYBMEP) Cu
//	wt]xH쎺	http://jsdiy.webcrow.jp

/*
yXVz
2017/06/18	v1.00	
	x[XƂAPI: BME280_driver-master [Bosch Sensortec (github)]
*/

#include <avr/io.h>
#include <util/delay.h>
#include "bme280lib.h"
#include "twilib.h"


//BME280̃X[uAhX(7bit)
#define	BME280_SLAVE_ADDRESS	0x76
/*
BME280SDO[qŃX[uAhXLSB߂邱ƂłB
GND  ƃV[gȂ'0'X[uAhX:0x76
VDDIOƃV[gȂ'1'X[uAhX:0x77
W['GYBMEP'ł0x76ƂȂĂB
ZT[3̃pbh͍ GND,SDO,VDDIO ŁAGND-SDÕp^[ȂĂB
p^[JbgSDO-VDDIOn_V[g0x77ɂA
@GND-SDOn_V[g0x76ɖ߂łB

vAbvR10kW[ɎρB
*/

//=========================================================
//	Calibration data
//=========================================================

//read calibration data: define
//(1) WX^calib00..calib25
#define CALPARAMS1_START_ADDR	0x88
#define CALPARAMS1_END_ADDR		0xA1

//(2) WX^calib26..calib41
#define CALPARAMS2_START_ADDR	0xE1
#define CALPARAMS2_END_ADDR		0xF0

//WX^press_msb..hum_lsb
#define SENSORVALUE_START_ADDR	0xF7
#define SENSORVALUE_END_ADDR	0xFE

#define BYTE_LENGTH(st,ed)	((ed) - (st) + 1)
#define MAX(a,b)	((a) < (b) ? (b) : (a))

#define CALPARAMS1_LENGTH	BYTE_LENGTH(CALPARAMS1_START_ADDR, CALPARAMS1_END_ADDR)
#define CALPARAMS2_LENGTH	BYTE_LENGTH(CALPARAMS2_START_ADDR, CALPARAMS2_END_ADDR)
#define CALPARAMS_LENGTH	MAX(CALPARAMS1_LENGTH, CALPARAMS2_LENGTH)

#define SENSORVALUE_LENGTH	BYTE_LENGTH(SENSORVALUE_START_ADDR, SENSORVALUE_END_ADDR)

//press,temp,hum̃WX^ǂݍݗpobt@
static	uint8_t	svbuf[SENSORVALUE_LENGTH];

//svbuf(oCgz)̃CfbNX
enum ESigByteBufIdx
{
	press_msb = 0,
	press_lsb,
	press_xlsb,
	temp_msb,
	temp_lsb,
	temp_xlsb,
	hum_msb,
	hum_lsb
};

//=========================================================
//	Compensation parameter
//=========================================================

//Zli[\
typedef	struct
{
	uint16_t	t1;
	int16_t		t2;
	int16_t		t3;
	uint16_t	p1;
	int16_t		p2;
	int16_t		p3;
	int16_t		p4;
	int16_t		p5;
	int16_t		p6;
	int16_t		p7;
	int16_t		p8;
	int16_t		p9;
	uint8_t		h1;
	int16_t		h2;
	uint8_t		h3;
	int16_t		h4;
	int16_t		h5;
	int8_t		h6;
	//
	int32_t	t_fine;
}
TBme280ComParam;

//ZT[Ǎl\lZoۂ̍Zl
static	TBme280ComParam	dig = {0};

//=========================================================
//	Control registers
//=========================================================

//Humidity oversampling
//Pressure oversampling
//Temperature oversampling
typedef	enum
{
	BME280_OSRS_Skipped			= 0b000,
	BME280_OSRS_Oversamplingx1	= 0b001,
	BME280_OSRS_Oversamplingx2	= 0b010,
	BME280_OSRS_Oversamplingx4	= 0b011,
	BME280_OSRS_Oversamplingx8	= 0b100,
	BME280_OSRS_Oversamplingx16	= 0b101
}
EBme280Osrs;

//sensor mode
typedef	enum
{
	BME280_MODE_SleepMode	= 0b00,
	BME280_MODE_ForcedMode	= 0b01,	//and 0b10
	BME280_MODE_NormalMode	= 0b11
}
EBme280SensorMode;

//inactive duration t_standby
typedef	enum
{
	BME280_TSB_1ms		= 0b000,	//0.5ms
	BME280_TSB_63ms		= 0b001,	//62.5ms
	BME280_TSB_125ms	= 0b010,
	BME280_TSB_250ms	= 0b011,
	BME280_TSB_500ms	= 0b100,
	BME280_TSB_1000ms	= 0b101,
	BME280_TSB_10ms		= 0b110,
	BME280_TSB_20ms		= 0b111
}
EBme280Tsb;

//time constant of the IIR filter
typedef	enum
{
	BME280_FILTER_FilterOff		= 0b000,
	BME280_FILTER_coefficient2	= 0b001,
	BME280_FILTER_coefficient4	= 0b010,
	BME280_FILTER_coefficient8	= 0b011,
	BME280_FILTER_coefficient16	= 0b100
}
EBme280Filter;

//Setting
typedef	struct
{
	//oversampling resolution
	EBme280Osrs	osrs_h;
	EBme280Osrs	osrs_p;
	EBme280Osrs	osrs_t;

	//sensor mode
	EBme280SensorMode	mode;
	
	//t_standby
	EBme280Tsb	t_sb;
	
	//filter coefficient
	EBme280Filter	filter;
}
TBme280Setting;

//ZT[̖ړIʐݒl
static	TBme280Setting settingParam = {0};

//=========================================================
//	vg^Cv錾
//=========================================================

static	void	WriteRegValue(uint8_t regAddr, uint8_t dataByte);
static	void	ReadRegValue(uint8_t startAddr, uint8_t readLength, uint8_t* retBuf);
static	void	DelayMSec(uint8_t msec);
static	uint8_t	ComCheck();
static	void	SoftReset();
static	void	ReadCalData();
static	void	WriteSensorParamToReg();
static	uint8_t	ComputeWaitTime();
static	uint16_t	GetModulo(int32_t num, uint16_t modorder);
static	int32_t		CompensateTemperature(int32_t ut);
static	uint32_t	CompensatePressure(uint32_t up);
static	uint32_t	CompensateHumidity(int32_t uh);

//=========================================================

//WX^AhXXXhɃf[^0xYY
static	void	WriteRegValue(uint8_t regAddr, uint8_t dataByte)
{
	TWI_Start();
	TWI_SlaveAddrW(BME280_SLAVE_ADDRESS);
	TWI_WriteByte(regAddr);
	TWI_WriteByte(dataByte);
	TWI_Stop();
}

//wAhXɊi[ꂽlǂ
static	void	ReadRegValue(uint8_t startAddr, uint8_t readLength, uint8_t* retBuf)
{
	uint8_t i;
	
	TWI_Start();
	TWI_SlaveAddrW(BME280_SLAVE_ADDRESS);
	TWI_WriteByte(startAddr);
	TWI_Stop();
	
	TWI_Start();
	TWI_SlaveAddrR(BME280_SLAVE_ADDRESS);
	for (i = 0; i < readLength; i++)
	{
		TWI_ReadByte((i == readLength - 1), retBuf + i);
	}
	TWI_Stop();
}

//1msPʂ̎ԑ҂
static	void	DelayMSec(uint8_t msec)
{
	uint8_t i;
	for (i = 0; i < msec; i++) { _delay_ms(1); }
}

//
uint8_t	BME280_Init()
{
	uint8_t	i, isOK;

	//ʐMm܂őҋ@iIDǍ܂ŌJԂƂɂ)
	for (i = 0, isOK = 0; i < 10 && !isOK; i++)
	{
		isOK = ComCheck();
		if (!isOK) { _delay_ms(10); }
	}
	if (!isOK) return 0;

	ReadCalData();
	return 1;
}

//ʐMmF
static	uint8_t	ComCheck()
{
	//WX^AhXD0hǂ0x60ʐMłĂ邱ƂɂȂ
	uint8_t	data;
	ReadRegValue(0xD0, 1, &data);
	
	return (data == 0x60);
}

//\tgZbg
static	void	SoftReset()
{
	//WX^AhXE0h0xB6
	WriteRegValue(0xE0, 0xB6);

	//uStart-up time: Max 2msv
	//Ƃ킯Ńn[hZbgƓԂ҂i{1ms]Tj
	_delay_ms(3);
}

//菀Ԃ擾
#if (0)	//ɕKvȂ
void	BME280_GetStatus(uint8_t* measuring, uint8_t* im_update)
{
	//WX^AhXF3hǂ
	uint8_t	status;
	ReadRegValue(0xF3, 1, &status);

	*measuring = (status >> 3) & 0x01;	//bit3
	*im_update = (status >> 0) & 0x01;	//bit0

	/*
	measuring:
	Automatically set to e1f whenever a conversion is running
	and back to e0f when the results have been transferred to the data registers.
	
	im_update:
	Automatically set to e1f when the NVM data are being copied to image registers
	and back to e0f when the copying is done.
	The data are copied at power-on-reset and before every conversion.
	*/
}
#endif

//read calibration data
static	void	ReadCalData()
{
	uint8_t buf[CALPARAMS_LENGTH];
	uint8_t i;
	
	//calibration dataǂ(1)
	ReadRegValue(CALPARAMS1_START_ADDR, CALPARAMS1_LENGTH, buf);

	//obt@̒lHč\̂֊i[
	i = 0;
	dig.t1 =			(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));		i += 2;
	dig.t2 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.t3 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.p1 =			(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));		i += 2;
	dig.p2 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.p3 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.p4 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.p5 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.p6 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.p7 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.p8 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.p9 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.h1 =	(uint8_t)buf[i];

	//calibration dataǂ(2)
	ReadRegValue(CALPARAMS2_START_ADDR, CALPARAMS2_LENGTH, buf);

	//obt@̒lHč\̂֊i[
	i = 0;
	dig.h2 =	(int16_t)(((uint16_t)buf[i]) | ((uint16_t)buf[i + 1] << 8));	i += 2;
	dig.h3 =	(uint8_t)buf[i];	i++;
	dig.h4 =	(int16_t)(((uint16_t)buf[i] << 4) | ((uint16_t)(buf[i + 1] & 0x0F)));
	dig.h5 =	(int16_t)(((uint16_t)((buf[i + 1] & 0xF0) >> 4)) | ((uint16_t)buf[i + 2] << 4));	i += 3;
	dig.h6 =	(int8_t)buf[i];
}

//f[^V[gɌfڂĂ邨߂̑p^[Zbg
/*	datasheet:
	5.5 Recommended modes of operation
	5.5.1 Weather monitoring
	5.5.2 Humidity sensing
	5.5.3 Indoor navigation
	5.5.4 Gaming
*/
void	BME280_PresetPattern(EBme280Pattern pattern)
{
	switch (pattern)
	{
	case BME280_PATTERN_WeatherMonitoring:
		settingParam.mode = BME280_MODE_ForcedMode;	//1 sample / minute
		settingParam.osrs_p = BME280_OSRS_Oversamplingx1;
		settingParam.osrs_t = BME280_OSRS_Oversamplingx1;
		settingParam.osrs_h = BME280_OSRS_Oversamplingx1;
		settingParam.filter = BME280_FILTER_FilterOff;
		break;

	case BME280_PATTERN_HumiditySensing:
		settingParam.mode = BME280_MODE_ForcedMode;	//1 sample / second
		settingParam.osrs_p = BME280_OSRS_Skipped;
		settingParam.osrs_t = BME280_OSRS_Oversamplingx1;
		settingParam.osrs_h = BME280_OSRS_Oversamplingx1;
		settingParam.filter = BME280_FILTER_FilterOff;
		break;

	case BME280_PATTERN_IndoorNavigation:
		settingParam.mode = BME280_MODE_NormalMode;
		settingParam.t_sb = BME280_TSB_1ms;		//T_standby = 0.5ms
		settingParam.osrs_p = BME280_OSRS_Oversamplingx16;
		settingParam.osrs_t = BME280_OSRS_Oversamplingx2;
		settingParam.osrs_h = BME280_OSRS_Oversamplingx1;
		settingParam.filter = BME280_FILTER_coefficient16;
		break;

	case BME280_PATTERN_Gaming:
		settingParam.mode = BME280_MODE_NormalMode;
		settingParam.t_sb = BME280_TSB_1ms;		//T_standby = 0.5ms
		settingParam.osrs_p = BME280_OSRS_Oversamplingx4;
		settingParam.osrs_t = BME280_OSRS_Oversamplingx1;
		settingParam.osrs_h = BME280_OSRS_Skipped;
		settingParam.filter = BME280_FILTER_coefficient16;
		break;

	default:
		break;
	}

	//WX^ɑp[^
	SoftReset();
	WriteSensorParamToReg();
}

//ZT[̐ݒlWX^ɏ
static	void	WriteSensorParamToReg()
{
	uint8_t	configRegValue, ctrlMeasRegValue, ctrlHumRegValue;
	uint8_t msec;

	configRegValue = ((settingParam.t_sb << 5) | (settingParam.filter << 2)) & 0xFC;	//LSB2bit00Ƃ
	WriteRegValue(0xF5, configRegValue);
	
	ctrlHumRegValue = settingParam.osrs_h & 0x07;
	WriteRegValue(0xF2, ctrlHumRegValue);	//ctrlMeasRegValueOɏނ
	
	ctrlMeasRegValue = ((settingParam.osrs_t << 5) | (settingParam.osrs_p << 2) | settingParam.mode) & 0xFF;
	WriteRegValue(0xF4, ctrlMeasRegValue);	//ctrlHumRegValueɏނ

	//p[^̃WX^񂾌Af[^ǂݍ߂܂ł̎ԑ҂
	msec = ComputeWaitTime();
	DelayMSec(msec);
}

//DELAY DEFINITIONS
#define	T_INIT_MAX				(20)	// 20/16 = 1.25 ms
#define	T_MEASURE_PER_OSRS_MAX	(37)	// 37/16 = 2.3125 ms
#define	T_SETUP_PRESSURE_MAX	(10)	// 10/16 = 0.625 ms
#define	T_SETUP_HUMIDITY_MAX	(10)	// 10/16 = 0.625 ms

//Computing waiting time for sensor data read
//Return value: The value of delay time for force mode
static	uint8_t	ComputeWaitTime()
{
	uint8_t	msec;

	msec = (T_INIT_MAX + T_MEASURE_PER_OSRS_MAX *
		(((1 << settingParam.osrs_t) >> 1) +
		((1 << settingParam.osrs_p) >> 1) +
		((1 << settingParam.osrs_h) >> 1)) +
		((settingParam.osrs_p > 0) ? T_SETUP_PRESSURE_MAX : 0) +
		((settingParam.osrs_h > 0) ? T_SETUP_HUMIDITY_MAX : 0) + 15) / 16;
	
	return msec;
}

//model擾
#if (0)	//gp
static	EBme280SensorMode	GetSensorMode()
{
	uint8_t	ctrlMeasReg, mode;
	EBme280SensorMode modes[] =
	{
		BME280_MODE_SleepMode,
		BME280_MODE_ForcedMode, BME280_MODE_ForcedMode,
		BME280_MODE_NormalMode
	};

	ReadRegValue(0xF4, 1, &ctrlMeasReg);
	mode = ctrlMeasReg & 0x03;
	return modes[mode];
}
#endif

//ZT[瑪l擾i\lɕϊ)
void	BME280_GetSensorValue(TBme280Value* retValue)
{
	int32_t ut, ct;
	uint32_t up, cp, uh, ch;

	//ForcedModȅꍇAmodeWX^ForcedModeށiSleepModeԂ瑪Ԃ֑Jڂj
	if (settingParam.mode == BME280_MODE_ForcedMode) { WriteSensorParamToReg(); }

	//read sensor value - Press, Temp, Hum
	ReadRegValue(SENSORVALUE_START_ADDR, SENSORVALUE_LENGTH, svbuf);

	//uncompensated {pressure, temperature, humidity} value
	up =		((uint32_t)svbuf[press_msb] << 12) | ((uint32_t)svbuf[press_lsb] << 4) | ((uint32_t)svbuf[press_xlsb] >> 4);
	ut = (int32_t)(((uint32_t)svbuf[temp_msb] << 12) | ((uint32_t)svbuf[temp_lsb] << 4) | ((uint32_t)svbuf[temp_xlsb] >> 4));
	uh =		((uint32_t)svbuf[hum_msb] << 8) | (uint32_t)svbuf[hum_lsb];

	//\lZo
	ct = CompensateTemperature(ut);
	retValue->temperature.integer = (int16_t)(ct / 100);
	retValue->temperature.decimal = (uint8_t)GetModulo(ct, 100);

	cp = CompensatePressure(up);
	retValue->pressure.integer = (int16_t)(cp / 100);
	retValue->pressure.decimal = (uint8_t)(cp % 100);
	
	ch = CompensateHumidity(uh);
	retValue->humidity.integer = (int16_t)(ch / 1024);
	retValue->humidity.decimal = (uint8_t)(((ch % 1024) * 1000) / 1024);
	retValue->humidity.decimal /= 10;	//3ʂ؂̂
}

//]߂iɑΉj
//	num:		̐
//		modorder:	10,100,1000,10000 ̂ꂩi_ȉ1-4߂邱Ƃɑj
static	uint16_t	GetModulo(int32_t num, uint16_t modorder)
{
	uint32_t	unum;
	uint16_t	modnum;

	unum = (num < 0) ? -num : num;
	modnum = (uint16_t)(unum % modorder);
	return modnum;
}

//Reads actual temperature from uncompensated temperature
//Output value of "5123" equals 51.23 DegC.
static	int32_t	CompensateTemperature(int32_t ut)
{
	int32_t	x1, x2;
	int32_t	temperature;

	// calculate x1
	x1 = ((((ut >> 3) - ((int32_t)dig.t1 << 1))) * ((int32_t)dig.t2)) >> 11;
	
	// calculate x2
	x2 = (((((ut >> 4) - ((int32_t)dig.t1)) * ((ut >> 4) - ((int32_t)dig.t1))) >> 12) * ((int32_t)dig.t3)) >> 14;
	
	// calculate t_fine
	//fine resolution temperature value over to the pressure compensation formula
	dig.t_fine = x1 + x2;

	// calculate temperature
	temperature = (dig.t_fine * 5 + 128) >> 8;
	return	temperature;
}

//Reads actual pressure from uncompensated pressure
//Output value of "96386" equals 96386 Pa = 963.86 hPa = 963.86 millibar
#define	BME280_INIT_VALUE	(0)
#define	BME280_INVALID_DATA	(0)

static	uint32_t	CompensatePressure(uint32_t up)
{
	int32_t x1 = BME280_INIT_VALUE;
	int32_t x2;
	uint32_t pressure;

	// calculate x1
	x1 = (((int32_t)dig.t_fine) >> 1) - (int32_t)64000;
	
	// calculate x2
	x2 = (((x1 >> 2) * (x1 >> 2)) >> 11) * ((int32_t)dig.p6);
	x2 = x2 + ((x1 * ((int32_t)dig.p5)) << 1);
	x2 = (x2 >> 2) + (((int32_t)dig.p4) << 16);
	
	// calculate x1
	x1 = (((dig.p3 * (((x1 >> 2) * (x1 >> 2)) >> 13)) >> 3) + ((((int32_t)dig.p2) * x1) >> 1)) >> 18;
	x1 = ((((32768 + x1)) * ((int32_t)dig.p1)) >> 15);

	// calculate pressure
	pressure = (((uint32_t)(((int32_t)1048576) - up) - (x2 >> 12))) * 3125;
	
	if (pressure < 0x80000000) {
		// Avoid exception caused by division by zero
		if (x1 != BME280_INIT_VALUE) {
			pressure = (pressure << 1) / ((uint32_t)x1);
		} else {
			return BME280_INVALID_DATA;
		}
	} else {
		// Avoid exception caused by division by zero
		if (x1 != BME280_INIT_VALUE) {
			pressure = (pressure / (uint32_t)x1) * 2;
		} else {
			return BME280_INVALID_DATA;
		}
	}

	x1 = (((int32_t)dig.p9) * ((int32_t)(((pressure >> 3) * (pressure >> 3)) >> 13))) >> 12;
	x2 = (((int32_t)(pressure >> 2)) * ((int32_t)dig.p8)) >> 13;
	
	pressure = (uint32_t)((int32_t)pressure + ((x1 + x2 + dig.p7) >> 4));
	return pressure;
}

//Reads actual humidity from uncompensated humidity
//Returns the value in %rH as unsigned 32bit integer in Q22.10 format(22 integer 10 fractional bits).
//An output value of 42313 represents 42313 / 1024 = 41.321 %rH
static	uint32_t	CompensateHumidity(int32_t uh)
{
	int32_t x1;
	uint32_t humidity;

	// calculate x1
	x1 = (dig.t_fine - ((int32_t)76800));
	x1 = (((((uh << 14) - (((int32_t)dig.h4) << 20) - (((int32_t)dig.h5) * x1)) +
		((int32_t)16384)) >> 15) * (((((((x1 * ((int32_t)dig.h6)) >> 10) *
		(((x1 * ((int32_t)dig.h3)) >> 11) + ((int32_t)32768))) >> 10) +
		((int32_t)2097152)) * ((int32_t)dig.h2) + 8192) >> 14));
	x1 = (x1 - (((((x1 >> 15) * (x1 >> 15)) >> 7) * ((int32_t)dig.h1)) >> 4));
	x1 = (x1 < 0 ? 0 : x1);
	x1 = (x1 > 419430400 ? 419430400 : x1);
	
	humidity = (uint32_t)(x1 >> 12);
	return humidity;
}
