/*****************************************************************************
*
* Atmel Corporation
*
* File              : USI_TWI_Master.h
* Compiler          : AVRGCC Toolchain version 3.4.2
* Revision          : $Revision: 992 $
* Date              : $Date: 2013-11-07 $
* Updated by        : $Author: Atmel $
*
* Support mail      : avr@atmel.com
*
* Supported devices : All device with USI module can be used.
*                     The example is written for the ATmega169, ATtiny26 and ATtiny2313
*
* AppNote           : AVR310 - Using the USI module as a TWI Master
*
* Description       : This is an implementation of an TWI master using
*                     the USI module as basis. The implementation assumes the AVR to
*                     be the only TWI master in the system and can therefore not be
*                     used in a multi-master system.
* Usage             : Initialize the USI module by calling the USI_TWI_Master_Initialise() 
*                     function. Hence messages/data are transceived on the bus using
*                     the USI_TWI_Start_Transceiver_With_Data() function. If the transceiver
*                     returns with a fail, then use USI_TWI_Get_Status_Info to evaluate the 
*                     couse of the failure.
*
****************************************************************************/
//	2017/06	初版
//	2017/09	usiTwiAvrDef.hを廃止し、AVRのポート定義をusiTwiMaster.hに取り込んだ。

#ifndef	_USI_TWI_MASTER_H_
#define	_USI_TWI_MASTER_H_


// Defines controlling timing limits
//	TWI_STANDARD_MODE	: TWI STANDARD mode timing limits. SCL <= 100kHz
//	TWI_FAST_MODE		: TWI FAST mode timing limits.     SCL  = 100-400kHz
//	↓どちらか一方を有効にし、他方をコメントアウトする。
//
#define	TWI_STANDARD_MODE
//#define	TWI_FAST_MODE

void	TWI_MasterInit();
uint8_t	TWI_Start();
uint8_t	TWI_SlaveAddrW(uint8_t sla);
uint8_t	TWI_SlaveAddrR(uint8_t sla);
uint8_t	TWI_WriteByte(uint8_t data);
uint8_t	TWI_ReadByte(uint8_t isLastOne, uint8_t* data);
uint8_t	TWI_Stop();

//AVRのポート定義
#if	defined(__AVR_ATtiny26__)
    #define DDR_USI             DDRB
    #define PORT_USI            PORTB
    #define PIN_USI             PINB
    #define PORT_USI_SDA        PORTB0
    #define PORT_USI_SCL        PORTB2
    #define PIN_USI_SDA         PINB0
    #define PIN_USI_SCL         PINB2
    #define USI_START_COND_INT  USISIF
    #define USI_START_VECTOR    USI_STRT_vect
    #define USI_OVERFLOW_VECTOR USI_OVF_vect
#endif

#if	defined(__AVR_ATtiny2313__) | \
	defined(__AVR_ATtiny2313A__) | defined(__AVR_ATtiny4313__)
    #define DDR_USI             DDRB
    #define PORT_USI            PORTB
    #define PIN_USI             PINB
    #define PORT_USI_SDA        PORTB5
    #define PORT_USI_SCL        PORTB7
    #define PIN_USI_SDA         PINB5
    #define PIN_USI_SCL         PINB7
    #define USI_START_COND_INT  USISIF
    #define USI_START_VECTOR    USI_START_vect
    #define USI_OVERFLOW_VECTOR USI_OVERFLOW_vect
#endif

#if	defined(__AVR_ATtiny44__) | defined(__AVR_ATtiny44A__) | \
	defined(__AVR_ATtiny84__)
	#define DDR_USI             DDRA
	#define PORT_USI            PORTA
	#define PIN_USI             PINA
	#define PORT_USI_SDA        PORTA6
	#define PORT_USI_SCL        PORTA4
	#define PIN_USI_SDA         PINA6
	#define PIN_USI_SCL         PINA4
	#define USI_START_COND_INT  USISIF
	#define USI_START_VECTOR    USI_STR_vect
	#define USI_OVERFLOW_VECTOR USI_OVF_vect
#endif

#if	defined(__AVR_ATtiny25__) | \
	defined(__AVR_ATtiny45__) | \
	defined(__AVR_ATtiny85__)
    #define DDR_USI             DDRB
    #define PORT_USI            PORTB
    #define PIN_USI             PINB
    #define PORT_USI_SDA        PB0
    #define PORT_USI_SCL        PB2
    #define PIN_USI_SDA         PINB0
    #define PIN_USI_SCL         PINB2
    #define USI_START_COND_INT  USISIF
    #define USI_START_VECTOR    USI_START_vect
    #define USI_OVERFLOW_VECTOR USI_OVF_vect
#endif

#if	defined(__AVR_ATmega165__) | defined(__AVR_ATmega169PA__) |\
	defined(__AVR_ATmega325__) | defined(__AVR_ATmega3250__) | \
	defined(__AVR_ATmega645__) | defined(__AVR_ATmega6450__) | \
	defined(__AVR_ATmega329__) | defined(__AVR_ATmega3290__) | \
	defined(__AVR_ATmega649__) | defined(__AVR_ATmega6490__)
    #define DDR_USI             DDRE
    #define PORT_USI            PORTE
    #define PIN_USI             PINE
    #define PORT_USI_SDA        PORTE5
    #define PORT_USI_SCL        PORTE4
    #define PIN_USI_SDA         PINE5
    #define PIN_USI_SCL         PINE4
    #define USI_START_COND_INT  USISIF
    #define USI_START_VECTOR    USI_START_vect
    #define USI_OVERFLOW_VECTOR USI_OVERFLOW_vect
#endif

#if	defined(__AVR_ATmega169__)
    #define DDR_USI             DDRE
    #define PORT_USI            PORTE
    #define PIN_USI             PINE
    #define PORT_USI_SDA        PORTE5
    #define PORT_USI_SCL        PORTE4
    #define PIN_USI_SDA         PINE5
    #define PIN_USI_SCL         PINE4
    #define USI_START_COND_INT  USISIF
    #define USI_START_VECTOR    USI_STRT_vect
    #define USI_OVERFLOW_VECTOR USI_OVF_vect
#endif


#endif	//_USI_TWI_MASTER_H_
